try:
    from cStringIO import StringIO as BytesIO
except ImportError:  # pragma: no cover
    # Python 3
    from io import BytesIO
import errno
import io
import os
import uuid
import requests
from sqlalchemy import schema
from sqlalchemy import types
from sqlalchemy.orm.session import object_session
from repoze.filesafe import create_file
from repoze.filesafe import open_file
from repoze.filesafe import delete_file
from s4u.sqlalchemy.meta import BaseObject
from s4u.image.util import extension_for_image_data
from s4u.image.scale import scale_image


def generate_path(extension):
    """Generate a filename within the image storage. The is based on a
    random string and the given extension. A three-level directory
    structure is used.
    """
    filename = '%s%s' % (uuid.uuid4(), extension)
    filename = os.path.join(filename[0], filename[1:3], filename)
    return filename


class Image(BaseObject):
    """A source image.
    """
    root_path = None
    __tablename__ = 'image'

    id = schema.Column(types.Integer(),
            schema.Sequence('image_id_seq', optional=True),
            primary_key=True, autoincrement=True)
    #: Relative filesystem path for the image
    path = schema.Column(types.String(128), unique=True)
    #: URL for (canonical) image.
    url = schema.Column(types.Text())

    def __init__(self, data=None, filename=None, url=None):
        if not (data is not None or url is not None):
            raise ValueError('You must provide at least one of data or url')
        self.url = url
        if data is not None:
            self._set_content(data, filename)

    @property
    def filesystem_path(self):
        """Return the (absolute) filesystem path for the image data."""
        if self.root_path is None:
            raise AttributeError('root_path not set')
        if not self.path:
            return None
        return os.path.join(self.root_path, self.path)

    def download(self, timeout=0.5):  # pragma: no cover
        """Download a remote image so we have a local copy.
        """
        if self.path is not None:
            raise TypeError('Image already has local data.')
        r = requests.get(self.url, timeout=timeout)
        extension = extension_for_image_data(BytesIO(r.content))
        self.path = generate_path(extension)
        file = create_file(self.filesystem_path, 'wb')
        if hasattr(file, 'fileno'):  # pragma: no cover (for testing only)
            try:
                os.fchmod(file.fileno(), 0o644)
            except io.UnsupportedOperation:  # BytesIO for testing
                pass
        file.write(r.content)
        file.close()

    def scale(self, width=None, height=None, crop=False,
            strip_whitespace=False):
        """Return a scaled version of this image. If a matching
        :py:class:`ImageScale` is found it is returned directly.  Otherwise the
        image will be scaled and a new ImageScale instance is created.

        See :py:func:`scale_image <s4u.image.scale.scale_image>` for more
        information on the scaling parameters.

        :rtype: :py:class:`ImageScale`
        """
        if not self.path:
            raise TypeError('Can not scale an image that is not stored locally.')

        if not (width or height):
            raise ValueError('You must specify either width or height')

        session = object_session(self)
        scale = session.query(ImageScale)\
                .filter(ImageScale.image_id == self.id)\
                .filter(ImageScale.param_width == (width or 0))\
                .filter(ImageScale.param_height == (height or 0))\
                .filter(ImageScale.param_crop == crop)\
                .filter(ImageScale.param_strip_whitespace == strip_whitespace)\
                .first()
        if scale is None:
            scale = ImageScale(self, width, height, crop, strip_whitespace)
            session.add(scale)
        return scale

    def delete(self):
        """Delete this image including all scales from the database and disk.
        """
        self._delete_scales()
        try:
            delete_file(self.filesystem_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        object_session(self).delete(self)

    def replace(self, data, filename=None):
        """Replace the image contents.

        This method should be used to modify images. This method will
        delete old old image scales and switch to a new filename and
        URL.

        :param str data: Raw image data.
        :param str filename: Image filename (optional)

        """
        self._delete_scales()
        try:
            delete_file(self.filesystem_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        self._set_content(data, filename)

    def _set_content(self, data, filename=None):
        extension = None
        if filename:
            extension = os.path.splitext(filename)[1]
        if not extension:
            extension = extension_for_image_data(BytesIO(data))
        self.path = generate_path(extension)
        file = create_file(self.filesystem_path, 'wb')
        if hasattr(file, 'fileno'):  # pragma: no cover (for testing only)
            try:
                os.fchmod(file.fileno(), 0o644)
            except io.UnsupportedOperation:  # BytesIO for testing
                pass
        file.write(data)
        file.close()

    def _delete_scales(self):
        """Remove all existing image scales.
        """
        session = object_session(self)
        scales = session.query(ImageScale).filter(ImageScale.image_id == self.id)
        for scale in scales:
            try:
                delete_file(scale.filesystem_path)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
        scales.delete()


class ImageScale(BaseObject):
    """A scaled version of image. Each :obj:`Image` can have many different
    scaled versions. The final dimensions of the image are stored to allow
    efficient creation of HTML ``img`` tags.
    """
    root_path = None
    __tablename__ = 'image_scale'
    __table_args__ = (
            schema.UniqueConstraint('image_id',
                                    'param_width', 'param_height',
                                    'param_crop', 'param_strip_whitespace'),
            schema.CheckConstraint('param_width>0 or param_height>0'),
            {})

    id = schema.Column(types.Integer(),
            schema.Sequence('image_scale_id_seq', optional=True),
            primary_key=True, autoincrement=True)
    path = schema.Column(types.String(128), nullable=False, unique=True)
    image_id = schema.Column(types.Integer(),
            schema.ForeignKey(Image.id,
                ondelete='CASCADE', onupdate='CASCADE'),
            nullable=False)
    param_width = schema.Column(types.Integer(), nullable=False)
    param_height = schema.Column(types.Integer(), nullable=False)
    param_crop = schema.Column(types.Boolean(), nullable=False)
    param_strip_whitespace = schema.Column(types.Boolean(), nullable=False)

    #: The width in pixels of the scaled image.
    width = schema.Column(types.Integer(), nullable=False)
    #: The heighy in pixels of the scaled image.
    height = schema.Column(types.Integer(), nullable=False)

    def __init__(self, image, width=None, height=None, crop=False,
            strip_whitespace=False):
        self.image_id = image.id
        self.param_width = width or 0
        self.param_height = height or 0
        self.param_crop = crop
        self.param_strip_whitespace = strip_whitespace

        file = open_file(image.filesystem_path, 'rb')
        (data, format, size) = scale_image(file, width, height, crop,
                strip_whitespace)
        self.path = generate_path('.' + format.lower())
        self.width = size[0]
        self.height = size[1]
        file = create_file(self.filesystem_path, 'wb')
        if hasattr(file, 'fileno'):  # pragma: no cover (for testing only)
            try:
                os.fchmod(file.fileno(), 0o644)
            except io.UnsupportedOperation:  # BytesIO for testing
                pass
        file.write(data)
        file.close()

    @property
    def filesystem_path(self):
        """Return the (absolute) filesystem path for the image data."""
        if self.root_path is None:
            raise AttributeError('root_path not set')
        return os.path.join(self.root_path, self.path)

    def _delete(self):
        """Remove image scale from database and filesystem.
        """
        try:
            delete_file(self.filesystem_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        object_session(self).delete(self)


schema.Index('image_scale_parameters',
             ImageScale.image_id,
             ImageScale.param_width,
             ImageScale.param_height,
             ImageScale.param_crop,
             ImageScale.param_strip_whitespace)

__all__ = ['Image', 'ImageScale']



import errno
import os
import unittest
from s4u.image.testing import ImageTestCase


class generate_path_tests(unittest.TestCase):
    def generate_path(self, *a, **kw):
        from s4u.image.model import generate_path
        return generate_path(*a, **kw)

    def test_three_levels(self):
        path = self.generate_path('.xys')
        self.assertEqual(len(path.split(os.path.sep)), 3)

    def test_keep_extension(self):
        path = self.generate_path('.xyz')
        self.assertTrue(path.endswith('.xyz'))
        self.assertTrue(not path.endswith('..xyz'))


class ImageTests(ImageTestCase):
    def Image(self, *a, **kw):
        from s4u.image.model import Image
        return Image(*a, **kw)

    def ImageScale(self, *a, **kw):
        from s4u.image.model import ImageScale
        return ImageScale(*a, **kw)

    def test_init_url_only(self):
        image = self.Image(url='http://media.example.com')
        self.assertEqual(image.url, 'http://media.example.com')
        self.assertEqual(image.path, None)

    def test_init_no_data_or_url(self):
        self.assertRaises(ValueError, self.Image)

    def test_set_content_extension_from_filename(self):
        from s4u.image.testing import PNG
        image = self.Image(PNG, 'test.gif')
        self.assertTrue(image.path.endswith('.gif'))

    def test_set_content_extension_from_data_if_filename_without_extension(self):
        from s4u.image.testing import PNG
        image = self.Image(PNG, 'test')
        self.assertTrue(image.path.endswith('.png'))

    def test_set_content_extension_from_data_if_no_filename(self):
        from s4u.image.testing import PNG
        image = self.Image(PNG)
        self.assertTrue(image.path.endswith('.png'))

    def test_filesystem_path_not_configured(self):
        from s4u.image.testing import PNG
        from s4u.image import configure
        image = self.Image(PNG)
        configure(None, None)  # Undo test setup
        self.assertRaises(AttributeError, getattr, image, 'filesystem_path')

    def test_filesystem_path_remote_image(self):
        image = self.Image(url='http://example.com/')
        self.assertEqual(image.filesystem_path, None)

    def test_download_already_local(self):
        from s4u.image.testing import PNG
        image = self.Image(PNG)
        self.assertRaises(TypeError, image.download)

    def test_download_png_image(self):
        from s4u.image.testing import PNG
        import mock
        image = self.Image(url='http://example.com/')
        response = mock.Mock()
        response.content = PNG
        with mock.patch('requests.get', return_value=response):
            image.download()
            self.assertTrue(image.path.endswith('.png'))

    def test_scale_no_size(self):
        from s4u.image.testing import PNG
        image = self.Image(PNG)
        self.assertRaises(ValueError, image.scale, 0, 0)

    def test_scale_remote_image_not_possible(self):
        image = self.Image(url='http://example.com/')
        self.assertRaises(TypeError, image.scale, 150)

    def test_scale_existing_scale(self):
        from s4u.sqlalchemy import meta
        from s4u.image.testing import PNG
        session = meta.Session()
        image = self.Image(PNG)
        session.add(image)
        session.flush()
        scale = self.ImageScale(image, width=123)
        session.add(scale)
        result = image.scale(width=123)
        self.assertTrue(result is scale)

    def test_scale_to_new_scale(self):
        from s4u.sqlalchemy import meta
        from s4u.image.testing import PNG
        from s4u.image.model import ImageScale
        session = meta.Session()
        image = self.Image(PNG)
        session.add(image)
        result = image.scale(width=123)
        self.assertTrue(isinstance(result, ImageScale))
        self.assertEqual(result.param_width, 123)
        self.assertEqual(result.param_height, 0)
        self.assertEqual(result.param_crop, False)

    def test_delete_missing_file(self):
        import mock
        from s4u.sqlalchemy.meta import Session
        from s4u.image.testing import JPEG
        from s4u.image.testing import PNG
        session = Session()
        image = self.Image(JPEG)
        session.add(image)
        session.flush()
        with mock.patch('s4u.image.model.delete_file',
                side_effect=OSError(errno.ENOENT, '')):
            image.delete()

    def test_delete_scales(self):
        import mock
        from s4u.sqlalchemy import meta
        from s4u.image.testing import PNG
        from s4u.image.model import ImageScale
        session = meta.Session()
        image = self.Image(PNG)
        session.add(image)
        scale = image.scale(width=123)
        session.flush()
        with mock.patch('s4u.image.model.delete_file') as delete_file:
            image._delete_scales()
            self.assertEqual(session.query(ImageScale).count(), 0)
            delete_file.assert_called_once_with(scale.filesystem_path)



class ImageScaleTests(ImageTestCase):
    def Image(self, *a, **kw):
        from s4u.image.model import Image
        return Image(*a, **kw)

    def ImageScale(self, *a, **kw):
        from s4u.image.model import ImageScale
        return ImageScale(*a, **kw)

    def test_init_extension_based_on_image_type(self):
        from s4u.image.testing import JPEG
        from s4u.image.testing import PNG
        image = self.Image(JPEG)
        scale = self.ImageScale(image, 10)
        self.assertTrue(scale.path.endswith('.jpeg'))
        image = self.Image(PNG)
        scale = self.ImageScale(image, 10)
        self.assertTrue(scale.path.endswith('.png'))

    def test_filesystem_path_not_configured(self):
        from s4u.image import configure
        from s4u.image.testing import PNG
        image = self.Image(PNG)
        scale = self.ImageScale(image, 1)
        configure(None, None)  # Undo test setup
        self.assertRaises(AttributeError, getattr, scale, 'filesystem_path')

    def test_delete(self):
        import mock
        from s4u.sqlalchemy.meta import Session
        from s4u.image.testing import JPEG
        from s4u.image.testing import PNG
        session = Session()
        image = self.Image(JPEG)
        session.add(image)
        session.flush()
        scale = self.ImageScale(image, 10)
        session.add(scale)
        session.flush()
        self.assertTrue(scale in session)
        with mock.patch('s4u.image.model.delete_file') as delete_file:
            scale._delete()
            self.assertTrue(scale in session.deleted)
            delete_file.assert_called_once_with(scale.filesystem_path)

    def test_delete_missing_file(self):
        import mock
        from s4u.sqlalchemy.meta import Session
        from s4u.image.testing import JPEG
        from s4u.image.testing import PNG
        session = Session()
        image = self.Image(JPEG)
        session.add(image)
        session.flush()
        scale = self.ImageScale(image, 10)
        session.add(scale)
        session.flush()
        self.assertTrue(scale in session)
        with mock.patch('s4u.image.model.delete_file',
                side_effect=OSError(errno.ENOENT, '')):
            scale._delete()



from sqlalchemy import schema
from sqlalchemy import types
from s4u.upgrade import upgrade_context
from s4u.upgrade import upgrade_step
from s4u.image import configure
from s4u.sqlalchemy import meta
from sqlalchemy.engine.reflection import Inspector
from . import _create_directories


@upgrade_context('image',
        [('--fs-images-original',
            {'type': str, 'required': True, 'dest': 'fs_images_original'}),
         ('--fs-images-scaled',
            {'type': str, 'required': True, 'dest': 'fs_images_scaled'})])
def setup_image_paths(options):
    configure(options.fs_images_original, options.fs_images_scaled)
    return {'fs.images.original': options.fs_images_original,
            'fs.images.scaled': options.fs_images_scaled}


@upgrade_step(require=['image'])
def create_directories(environment):
    _create_directories(environment['fs.images.original'])
    _create_directories(environment['fs.images.scaled'])


@upgrade_step(require=['sql'])
def add_missing_entities(environment):  # pragma: no cover
    engine = environment['sql-engine']
    meta.metadata.create_all(engine)


@upgrade_step(require=['sql'])
def add_url_column(environment):
    engine = environment['sql-engine']
    alembic = environment['alembic']
    inspector = Inspector.from_engine(engine)
    columns = set(c['name']
                  for c in inspector.get_columns('image'))
    if 'url' not in columns:
        alembic.add_column('image', schema.Column('url', types.Text()))



try:
    from cStringIO import StringIO as BytesIO
except ImportError:  # pragma: no cover
    # Python 3
    from io import BytesIO
import PIL.Image
import PIL.ImageChops
import PIL.ImageFile

# Set a larger buffer size. This fixes problems with jpeg decoding.
# See http://mail.python.org/pipermail/image-sig/1999-August/000816.html for
# details.
PIL.ImageFile.MAXBLOCK = 1000000


def scale_image(image, width=None, height=None, crop=False,
        strip_whitespace=False):
    """Scale the given image data to another size and return the result
    as a string or optionally write in to the file-like `result` object.

    :param file image: open file for image to scale
    :param int width: desired maximum image width
    :param int height: desired maximum image width
    :param bool crop: allow cropping image to fill full width and height
    :param bool strip_whitespace: crop surrounding whitespace before
      processing the image

    The return value is a tuple with the new image, the image format and
    a size-tuple.

    The `width`, `height`, `direction` parameters will be passed to
    :meth:`scale_pil_image`, which performs the actual scaling.
    """
    image = PIL.Image.open(image)

    # When we create a new image during scaling we loose the format
    # information, so remember it here.
    format = image.format
    if format == 'GIF':
        format = 'PNG'
    elif format != 'PNG':
        format = 'JPEG'

    if strip_whitespace:
        image = crop_surrounding_whitespace(image)

    image = scale_pil_image(image, width, height, crop)
    output = BytesIO()
    image.save(output, format, optimize=True)
    return (output.getvalue(), format, image.size)


def correct_colour_mode(image):
    """Make sure an image uses a colour handling scheme which allows for
    high quality scaling and can be rendered by web browsers.
    """
    if image.mode == '1':
        # Convert black&white to grayscale
        image = image.convert("L")
    elif image.mode == 'P':
        # Convert palette based images to 3x8bit+alpha
        image = image.convert('RGBA')
    elif image.mode == 'CMYK':
        # Convert CMYK to RGB since web browser can not handle CMYK
        image = image.convert('RGB')
    return image


def crop_surrounding_whitespace(image):
    """Remove surrounding empty space around an image.

    This implemenation assumes that the surrounding space has the same colour
    as the top leftmost pixel.

    :param image: PIL image
    :rtype: PIL image
    """
    bg = PIL.Image.new(image.mode, image.size, image.getpixel((0, 0)))
    diff = PIL.ImageChops.difference(image, bg)
    bbox = diff.getbbox()
    if not bbox:
        return image

    # XXX: if we want to crop the final image we should try to match the
    # desired aspect ratio here to make sure the last step will not
    # crop out real image data.
    return image.crop(bbox)


def center_crop(image, width, height):
    """Crop an image to the desired width and height. The crop is
    made from the middle of the image.

    :param image: PIL image
    :param int width: maximum width, or None of width is unrestrained
    :param int height: maximum height, or None of height is unrestrained
    :rtype: PIL image
    """
    (current_width, current_height) = image.size

    if width is not None and current_width > width:
        left = int((current_width - width) / 2.0)
        right = left + width
    else:
        left = 0
        right = current_width

    if height is not None and current_height > height:
        top = int((current_height - height) / 2.0)
        bottom = top + height
    else:
        top = 0
        bottom = current_height

    return image.crop((left, top, right, bottom))


def scale_pil_image(image, width=None, height=None, crop=False):
    """Scale a PIL image to another size.

    :param image: PIL Image instance
    :param int width: desired maximum image width
    :param int height: desired maximum image width
    :param bool crop: allow cropping image to fill full width and height
    :rtype: PIL Image

    The generated image is a JPEG image, unless the original is a GIF or PNG
    image. This is needed to make sure alpha channel information is not lost,
    which JPEG does not support.
    """
    if width is None and height is None:
        raise ValueError('Either width or height need to be given')

    image = correct_colour_mode(image)

    (current_width, current_height) = image.size

    scale_height = (float(height) / float(current_height)) \
            if height is not None else None
    scale_width = (float(width) / float(current_width)) \
            if width is not None else None

    if scale_height == scale_width:
        # The original already has the right aspect ratio, so use a fast
        # thumbnail to scale.
        image.thumbnail((width, height), PIL.Image.ANTIALIAS)
        return image

    if crop:
        scale = max(scale_width, scale_height)
    else:
        scale = min(filter(None, [scale_width, scale_height]))

    # Skip scale multiplication if possible to prevent off-by-one errors
    new_width = width if scale == scale_width \
            else int(current_width * scale)
    new_height = height if scale == scale_height \
            else int(current_height * scale)

    image.draft(image.mode, (new_width, new_height))
    image = image.resize((new_width, new_height), PIL.Image.ANTIALIAS)

    if not crop:
        return image

    return center_crop(image, width, height)



import tempfile
from repoze.filesafe.testing import setupDummyDataManager
from repoze.filesafe.testing import cleanupDummyDataManager
import s4u.sqlalchemy.testing
from s4u.image import configure


GIF = b'GIF89a\x01\x00\x01\x00\x00\x00\x00,\x00\x00\x00\x00\x01\x00\x01' \
      b'\x00\x00\x02\x02L\x01\x00;'

PNG = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00' \
      b'\x01\x08\x06\x00\x00\x00\x1f\x15\xc4\x89\x00\x00\x00\nIDATx\x9cc' \
      b'\x00\x01\x00\x00\x05\x00\x01\r\n-\xb4\x00\x00\x00\x00IEND\xaeB`\x82'

JPEG = b'\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x00\x00\x01\x00\x01\x00' \
       b'\x00\xff\xdb\x00C\x00\x08\x06\x06\x07\x06\x05\x08\x07\x07\x07\t\t' \
       b'\x08\n\x0c\x14\r\x0c\x0b\x0b\x0c\x19\x12\x13\x0f\x14\x1d\x1a\x1f' \
       b'\x1e\x1d\x1a\x1c\x1c $.\' ",#\x1c\x1c(7),01444\x1f\'9=82<.342\xff' \
       b'\xc0\x00\x0b\x08\x00\x01\x00\x01\x01\x01\x11\x00\xff\xc4\x00\x14' \
       b'\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
       b'\x00\x08\xff\xc4\x00\x14\x10\x01\x00\x00\x00\x00\x00\x00\x00\x00' \
       b'\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x08\x01\x01\x00\x00' \
       b'?\x00?\xbf\xff\xd9'


class ImageTestCase(s4u.sqlalchemy.testing.DatabaseTestCase):
    """Base class for tests dealing with image models. This base class
    sets up a database with the image tables and configures a
    dummy datamanager so image data does not have to be stored on the
    filesystem.
    """
    def setUp(self):
        super(ImageTestCase, self).setUp()
        setupDummyDataManager()
        tmp = tempfile.gettempdir()
        configure(tmp, tmp)

    def tearDown(self):
        cleanupDummyDataManager()
        super(ImageTestCase, self).tearDown()
        configure(None, None)

__all__ = ['GIF', 'PNG', 'JPEG', 'ImageTestCase']



import os
from s4u.image.model import Image
from s4u.image.model import ImageScale


def configure(original_path, scale_path):
    """Configure the filesystem paths used to store image files.

    :param original_path: filesystem path used for full size images.
    :param scale_path: filesystem path used for scaled images.
    """
    Image.root_path = original_path
    ImageScale.root_path = scale_path


def includeme(config):
    """Configure s4u.image using a Pyramid :py:class:`Configurator
    <pyramid.config.Configurator>` object. This will take the filesystem
    paths from the application settings using the keys
    ``fs.images.original`` and ``fs.images.scaled``.
    """
    settings = config.registry.settings
    configure(settings['fs.images.original'],
              settings['fs.images.scaled'])
    for key in ['original', 'scaled']:
        base_url = settings.get('fs.images.%s.url' % key)
        if base_url:
            config.add_static_view(base_url,
                    settings['fs.images.%s' % key])
        else:
            config.add_static_view('image-%s' % key,
                    settings['fs.images.%s' % key],
                    cache_max_age=86400 * 31)


def initialise_filesystem():
    """Create required filesystem.

    This function requires that s4u.image is already configured.

    It is safe to call this function multiple times: it will notice if a
    directory already exists.
    """
    _create_directories(Image.root_path)
    _create_directories(ImageScale.root_path)


def _create_directories(root_path):
    if not os.path.isdir(root_path):
        raise RuntimeError('%s is not a valid image root path' % root_path)
    for top in range(16):
        top_path = os.path.join(root_path, '%x' % top)
        if not os.path.exists(top_path):
            os.mkdir(top_path)
        for sub in range(256):
            sub_path = os.path.join(top_path, '%02x' % sub)
            if not os.path.exists(sub_path):
                os.mkdir(sub_path)



import unittest
from s4u.sqlalchemy.testing import DatabaseTestCase


class setup_image_paths_tests(unittest.TestCase):
    def setup_image_paths(self, *a, **kw):
        from s4u.image.upgrade import setup_image_paths
        return setup_image_paths(*a, **kw)

    def test_configure_paths(self):
        import mock
        with mock.patch('s4u.image.upgrade.configure') as mock_configure:
            options = DummyOptions(
                    fs_images_original='/tmp/foo',
                    fs_images_scaled='/tmp/buz')
            self.setup_image_paths(options)
            mock_configure.assert_called_once_with('/tmp/foo', '/tmp/buz')

    def test_environment(self):
        import mock
        with mock.patch('s4u.image.upgrade.configure'):
            options = DummyOptions(
                    fs_images_original='/tmp/foo',
                    fs_images_scaled='/tmp/buz')
            environment = self.setup_image_paths(options)
            self.assertEqual(
                    environment,
                    {'fs.images.original': '/tmp/foo',
                     'fs.images.scaled': '/tmp/buz'})


class create_directories_tests(unittest.TestCase):
    def create_directories(self, *a, **kw):
        from s4u.image.upgrade import create_directories
        return create_directories(*a, **kw)

    def test_create_original_directories(self):
        import mock
        with mock.patch('s4u.image.upgrade._create_directories') as mock_crea:
            self.create_directories({'fs.images.original': '/tmp/foo',
                                     'fs.images.scaled': '/tmp/buz'})
            mock_crea.assert_has_calls([mock.call('/tmp/foo')], True)

    def test_create_scaled_directories(self):
        import mock
        with mock.patch('s4u.image.upgrade._create_directories') as mock_crea:
            self.create_directories({'fs.images.original': '/tmp/foo',
                                     'fs.images.scaled': '/tmp/buz'})
            mock_crea.assert_has_calls([mock.call('/tmp/buz')], True)


class int_create_directories_tests(unittest.TestCase):
    def _create_directories(self, *a, **kw):
        from s4u.image.upgrade import _create_directories
        return _create_directories(*a, **kw)

    def test_all_dirs_exist(self):
        import mock
        with mock.patch('os.path.exists') as mock_exists:
            with mock.patch('os.mkdir') as mock_mkdir:
                mock_exists.return_value = True
                self._create_directories('/tmp')
                self.assertEqual(mock_exists.call_count, 16 + (16 * 256))
                self.assertEqual(mock_mkdir.called, False)

    def test_toplevel_missing(self):
        import mock
        with mock.patch('os.path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertRaises(RuntimeError, self._create_directories, '/')

    def test_create_all(self):
        import mock
        with mock.patch('os.path.exists') as mock_exists:
            with mock.patch('os.mkdir') as mock_mkdir:
                mock_exists.return_value = False
                self._create_directories('/tmp')
                self.assertEqual(mock_exists.call_count, 16 + (16 * 256))
                self.assertEqual(mock_mkdir.call_count, 16 + (16 * 256))



class add_url_column_tests(DatabaseTestCase):
    create_tables = False

    def add_url_column(self, *a, **kw):
        from s4u.image.upgrade import add_url_column
        return add_url_column(*a, **kw)

    def test_already_in_desired_situation(self):
        from s4u.sqlalchemy import meta
        import s4u.image.model
        s4u.image.model  # Hello PyFlakes
        meta.metadata.create_all(self.engine)
        self.add_url_column({'sql-engine': self.engine, 'alembic': None})

    def test_add_url_column(self):
        import mock
        engine = self.engine
        engine.execute(
                'CREATE TABLE image (id INT)')
        alembic = mock.Mock()
        self.add_url_column({'sql-engine': engine, 'alembic': alembic})
        self.assertEqual(alembic.add_column.call_count, 1)
        arguments = alembic.add_column.call_args
        self.assertEqual(arguments[0][0], 'image')
        self.assertEqual(arguments[0][1].name, 'url')





class DummyOptions(object):
    def __init__(self, **kw):
        for (k, v) in kw.items():
            setattr(self, k, v)



import unittest


class configure_tests(unittest.TestCase):
    def configure(self, *a, **kw):
        from s4u.image import configure
        return configure(*a, **kw)

    def test_set_original_path(self):
        from s4u.image.model import Image
        self.assertEqual(Image.root_path, None)
        try:
            self.configure('/tmp', None)
            self.assertEqual(Image.root_path, '/tmp')
        finally:
            Image.root_path = None

    def test_set_scale_path(self):
        from s4u.image.model import ImageScale
        self.assertEqual(ImageScale.root_path, None)
        try:
            self.configure(None, '/tmp')
            self.assertEqual(ImageScale.root_path, '/tmp')
        finally:
            ImageScale.root_path = None


class includeme_tests(unittest.TestCase):
    def includeme(self, *a, **kw):
        from s4u.image import includeme
        return includeme(*a, **kw)

    def test_set_original_path(self):
        from s4u.image.model import Image
        config = MockConfig({'fs.images.original': '/tmp',
                             'fs.images.scaled': None})
        self.assertEqual(Image.root_path, None)
        self.includeme(config)
        self.assertEqual(Image.root_path, '/tmp')

    def test_set_scale_path(self):
        from s4u.image.model import ImageScale
        config = MockConfig({'fs.images.original': None,
                             'fs.images.scaled': '/tmp'})
        self.assertEqual(ImageScale.root_path, None)
        self.includeme(config)
        self.assertEqual(ImageScale.root_path, '/tmp')

    def test_setup_static_views(self):
        config = MockConfig({'fs.images.original': '/fs/original',
                             'fs.images.scaled': '/fs/scaled'})
        self.includeme(config)
        self.assertEqual(len(config.static_views), 2)
        self.assertEqual(config.static_views[0]['path'], '/fs/original')
        self.assertTrue(config.static_views[0]['cache_max_age'] > 86400)
        self.assertEqual(config.static_views[1]['path'], '/fs/scaled')
        self.assertTrue(config.static_views[1]['cache_max_age'] > 86400)

    def test_setup_static_views_with_external_url(self):
        config = MockConfig({'fs.images.original': '/fs/original',
                             'fs.images.scaled': '/fs/scaled',
                             'fs.images.scaled.url': 'http://media.com/'})
        self.includeme(config)
        self.assertEqual(len(config.static_views), 2)
        self.assertEqual(config.static_views[0]['path'], '/fs/original')
        self.assertEqual(config.static_views[1]['name'], 'http://media.com/')
        self.assertEqual(config.static_views[1]['path'], '/fs/scaled')
        self.assertTrue('cache_max_age' not in config.static_views[1])


class MockRegistry:
    pass


class MockConfig:
    def __init__(self, settings):
        self.registry = MockRegistry()
        self.registry.settings = settings
        self.static_views = []

    def add_static_view(self, name, path, **kw):
        info = kw
        info['name'] = name
        info['path'] = path
        self.static_views.append(info)



import PIL.Image


def extension_for_image_data(data):
    '''Determine a suitable filesystem extension for image data. The extension
    is based on the image format.

    :param data: either a filename for an image file, or an open file handle.
    :return: extension including leading dot
    '''
    try:
        image = PIL.Image.open(data)
        image.load()
    except (IOError, OverflowError):
        raise ValueError('Not a valid image')

    return '.%s' % image.format.lower()



from setuptools import setup
from setuptools import find_packages

version = '2.3.2'

requires = [
    'setuptools',
    's4u.sqlalchemy >=1.2',
    'repoze.filesafe >=2.0',
    'Pillow',
    'requests',
    ]

test_requires = [
    'mock >=0.8',
    's4u.upgrade',
    ]

setup(name='s4u.image',
      version=version,
      description='2Style4You image library',
      long_description=open('README.rst').read(),
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Framework :: Pyramid',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Topic :: Software Development :: Libraries :: Python Modules',
        ],
      author='Simplon B.V. - Wichert Akkerman',
      author_email='wichert@wiggy.net',
      url='https://s4uimage.readthedocs.org/en/latest/',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      namespace_packages=['s4u'],
      zip_safe=False,
      install_requires=requires,
      tests_require=test_requires,
      extras_require={
          'docs': ['sphinx'],
          'tests': test_requires,
      },
      test_suite='s4u.image')



import unittest
import os
try:
    from cStringIO import StringIO as BytesIO
except ImportError:  # pragma: no cover
    # Python 3
    from io import BytesIO

HERE = os.path.dirname(__file__)


class scale_image_test(unittest.TestCase):
    def scale_image(self, *a, **kw):
        from s4u.image.scale import scale_image
        return scale_image(*a, **kw)

    def test_return_value(self):
        from s4u.image.testing import PNG
        result = self.scale_image(BytesIO(PNG), width=10)
        self.assertEqual(result[1], 'PNG')
        self.assertEqual(result[2], (10, 10))

    def test_gif_converted_to_png(self):
        from s4u.image.testing import GIF
        self.assertEqual(
            self.scale_image(BytesIO(GIF), width=10)[1], 'PNG')

    def test_jpeg_stays_jpeg(self):
        from s4u.image.testing import JPEG
        self.assertEqual(
            self.scale_image(BytesIO(JPEG), width=10)[1], 'JPEG')

    def test_bad1(self):
        image = open(os.path.join(HERE, 'bad1.jpg'), 'rb')
        (image, format, size) = self.scale_image(image, 200, 200, True, True)

    def test_bad2(self):
        image = open(os.path.join(HERE, 'bad2.jpg'), 'rb')
        (image, format, size) = self.scale_image(image, 200, 200, False, True)

    def test_bad3(self):
        image = open(os.path.join(HERE, 'bad3.jpg'), 'rb')
        (image, format, size) = self.scale_image(image, 200, 200, True, True)


class correct_colour_mode_tests(unittest.TestCase):
    def correct_colour_mode(self, *a, **kw):
        from s4u.image.scale import correct_colour_mode
        return correct_colour_mode(*a, **kw)

    def test_convert_black_white_to_greyscale(self):
        import PIL.Image
        self.assertEqual(
                self.correct_colour_mode(PIL.Image.new('1', (50, 50))).mode,
                'L')

    def test_convert_palette_to_rgba(self):
        import PIL.Image
        self.assertEqual(
                self.correct_colour_mode(PIL.Image.new('P', (50, 50))).mode,
                'RGBA')

    def test_convert_cmyk(self):
        import PIL.Image
        self.assertEqual(
                self.correct_colour_mode(PIL.Image.new('CMYK', (50, 50))).mode,
                'RGB')

    def test_rgb_image_not_touched(self):
        import PIL.Image
        image = PIL.Image.new('RGB', (50, 50))
        self.assertTrue(self.correct_colour_mode(image) is image)


class center_crop_tests(unittest.TestCase):
    def center_crop(self, *a, **kw):
        from s4u.image.scale import center_crop
        return center_crop(*a, **kw)

    def test_image_smaller_than_crop_size(self):
        import PIL.Image
        image = PIL.Image.new('L', (50, 60))
        crop = self.center_crop(image, 100, 100)
        self.assertEqual(crop.size, (50, 60))

    def test_crop_width_only(self):
        import PIL.Image
        image = PIL.Image.new('L', (50, 60))
        crop = self.center_crop(image, 30, 100)
        self.assertEqual(crop.size, (30, 60))

    def test_crop_height_only(self):
        import PIL.Image
        image = PIL.Image.new('L', (50, 60))
        crop = self.center_crop(image, 100, 30)
        self.assertEqual(crop.size, (50, 30))

    def test_center_horizontally(self):
        # Test by drawing a centered red rectangle, and checking there is
        # no black after cropping.
        import PIL.Image
        import PIL.ImageDraw
        image = PIL.Image.new('L', (100, 100))
        draw = PIL.ImageDraw.Draw(image)
        draw.rectangle((25, 0, 76, 100), fill='#ff0000')
        self.assertNotEqual(image.getcolors(), [(5000, 76)])
        crop = self.center_crop(image, 50, 100)
        self.assertEqual(crop.getcolors(), [(5000, 76)])

    def test_center_vertically(self):
        # Test by drawing a centered red rectangle, and checking there is
        # no black after cropping.
        import PIL.Image
        import PIL.ImageDraw
        image = PIL.Image.new('L', (100, 100))
        draw = PIL.ImageDraw.Draw(image)
        draw.rectangle((0, 25, 100, 76), fill='#ff0000')
        self.assertNotEqual(image.getcolors(), [(5000, 76)])
        crop = self.center_crop(image, 100, 50)
        self.assertEqual(crop.getcolors(), [(5000, 76)])


class scale_pil_image_tests(unittest.TestCase):
    def scale_pil_image(self, *a, **kw):
        from s4u.image.scale import scale_pil_image
        return scale_pil_image(*a, **kw)

    def _make_image(self, width, height):
        import PIL.Image
        return PIL.Image.new('L', (width, height))

    def test_width_or_height_required(self):
        self.assertRaises(ValueError, self.scale_pil_image, None, None, None)

    def test_scale_to_width_only(self):
        image = self._make_image(50, 60)
        scale = self.scale_pil_image(image, width=25)
        self.assertEqual(scale.size, (25, 30))

    def test_scale_to_height_only(self):
        image = self._make_image(50, 30)
        scale = self.scale_pil_image(image, height=60)
        self.assertEqual(scale.size, (100, 60))

    def test_scale_with_same_aspect_ratio(self):
        image = self._make_image(50, 50)
        scale = self.scale_pil_image(image, width=25, height=25, crop=False)
        self.assertEqual(scale.size, (25, 25))

    def test_scale_overly_wide_image_without_crop(self):
        image = self._make_image(100, 10)
        scale = self.scale_pil_image(image, width=30, height=30, crop=False)
        self.assertEqual(scale.size, (30, 3))

    def test_scale_overly_wide_image_with_crop(self):
        image = self._make_image(100, 10)
        scale = self.scale_pil_image(image, width=30, height=30, crop=True)
        self.assertEqual(scale.size, (30, 30))

    def test_scale_overly_tall_image_without_crop(self):
        image = self._make_image(10, 100)
        scale = self.scale_pil_image(image, width=30, height=30, crop=False)
        self.assertEqual(scale.size, (3, 30))

    def test_scale_overly_tall_image_with_crop(self):
        image = self._make_image(10, 100)
        scale = self.scale_pil_image(image, width=30, height=30, crop=True)
        self.assertEqual(scale.size, (30, 30))



import unittest
try:
    from cStringIO import StringIO as BytesIO
except ImportError:  # pragma: no cover
    # Python 3
    from io import BytesIO


class extension_for_image_data_tests(unittest.TestCase):
    def extension_for_image_data(self, *a, **kw):
        from s4u.image.util import extension_for_image_data
        return extension_for_image_data(*a, **kw)

    def test_invalid_image(self):
        input = BytesIO(b'invalid')
        self.assertRaises(ValueError,
                self.extension_for_image_data, input)

    def test_minimal_gif(self):
        from s4u.image.testing import GIF
        input = BytesIO(GIF)
        self.assertEqual(self.extension_for_image_data(input), '.gif')

    def test_minimal_png(self):
        from s4u.image.testing import PNG
        input = BytesIO(PNG)
        self.assertEqual(self.extension_for_image_data(input), '.png')
